/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/

package javax.tools.diagnostics.vm;

import javax.tools.diagnostics.vm.spi.DumpInitiatorCapabilities;

/**
 * Interface that describes the required capabilities of specific
 * dump initiator.
 * 
 * Classes that implement this interface can be called by the
 * Dump class to trigger a specific dump.
 *  
 */
public interface DumpInitiatorDelegate {

	/**
	 * Returns an object that describes the capabilities offered by
	 * this initiator
	 * 
	 * @return a populated capabilities object
	 */
	public DumpInitiatorCapabilities getCapabilities();
	
	
	/**
	 * Called to check that the delegate has all available resources
	 * to proceed.   
	 * This method should only be called by the Dump class
	 * This method will always be called before the first call to {@link dump}
	 * 
	 * @return true if the delegate is available 
	 */
	public  boolean available();

	/**
	 * Returns a Dump handle for a default configured dump
	 * 
	 * @return valid dump handle
	 */
	public DumpHandle createDumpHandle();
	
	/**
	 * Returns a Dump handle for a dump that will match the provided 
	 * descriptor on a "best can do" basis. 
	 * 
	 * @return valid dump handle
	 */
	public DumpHandle createDumpHandle(DumpDescriptor descriptor);
	
	
	/** 
	 * Returns a literal that describes the dump being produced by this
	 * initiator.  The value can be used by callers to the DumpFactory API
	 * to retrieve a specific dump initiator.
	 * 
	 * @return dump type literal
	 */
	public String getDumpType();
	 
}
