/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package javax.tools.diagnostics;

import java.io.File;
import java.io.IOException;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import javax.imageio.spi.ServiceRegistry;
import javax.tools.diagnostics.image.Image;
import javax.tools.diagnostics.image.ImageFactory;
import javax.tools.diagnostics.runtime.java.JavaRuntime;


/**
 * <p>
 * Experimental addition to the API.
 * <p>
 * This class provides a central registry for image factories
 * <p>
 * Image factories can be registered directly using the {@link #addFactory()} method.
 * <p>
 * The default registry obtained by calling {@link #getDefaultRegistry()} uses {@link  javax.imageio.spi.ServiceRegistry} to discover
 * ImageFactory implementations.
 * <p>
 * To register an ImageFactory implementation that can be discovered by the registry do the
 * following :
 * <ul> 
 * <li>Create a 'services' directory as a child of the manfest directory 'META-INF'
 * <li>Within this 'services' directory create a text file called 'javax.tools.diagnostics.image.ImageFactory'
 * <li>This text file should contain a single line which is the package qualified name of the ImageFactory 
 * implementation to be registered
 * </ul>  
 * </p>
 * @experimental
 *
 */
public class FactoryRegistry implements Iterable<ImageFactory>{

	private static FactoryRegistry defaultRegistry=null;
	private Set<ImageFactory> factories=new HashSet<ImageFactory>();
	
	/**
	 * <p>
	 * Creates an empty registry.  To obtain an populated registry use the {@link #getDefaultRegistry()} method
	 */
	public FactoryRegistry() {
		
	}
	
	/**
	 * <p>
	 * Returns the default registry.  This registy is preloaded with
	 * ImageFactory implementations discovered using the {@link  javax.imageio.spi.ServiceRegistry} 
	 * 
	 * @return default image factory
	 */
	public synchronized static FactoryRegistry getDefaultRegistry() {
		
		if(defaultRegistry==null) {
			defaultRegistry=new FactoryRegistry();
			defaultRegistry.addLocalFactories();
			
			
		}
		
		return defaultRegistry;
	}

	
	/**
	 * <p>
	 * Using {@link  javax.imageio.spi.ServiceRegistry} this method
	 * finds ImageFactory implementations and adds them to this instance
	 */
	private void addLocalFactories() {
		
	
		
		Iterator<ImageFactory> i= ServiceRegistry.lookupProviders(javax.tools.diagnostics.image.ImageFactory.class);
	 
		while(i.hasNext()) {
		   ImageFactory localFactory=i.next();
		   addFactory(localFactory);
		}
		
	}

	/**
	 * <p>
	 * Returns an {@link java.util.Iterator} of ImageFactories registered
	 * to this registry. 
	 */
	public Iterator<ImageFactory> iterator() {
		return factories.iterator();
	}

	/**
	 * <p>
	 * Adds an ImageFactory to the registry instance
	 * If the factory instance is already in the 
	 * registry it is not added again.
	 *  
	 * @param factory factory to add to registry
	 * @throws IllegalArgument if factory is null
	 * @return true if factory added 
	 */
	public boolean addFactory(ImageFactory factory) {
		if(factory==null) throw new IllegalArgumentException("factory parameter is null");
		return factories.add(factory);
		
	}

	/**
	 * <p>
	 * Always returns an array even if the registry is empty
	 * 
	 * @return Returns the factories in the registry as an array
	 * 
	 */
	public ImageFactory[] getFactories() {
		
		return factories.toArray(new ImageFactory[0]);
	}
	
	/**
	 * <p>
	 * Returns an appropriate {@link javax.tools.diagnostics.image.Image} for the
	 * provide file by locating the first registered image factory that can handle
	 * the case insensitive file name extension of the provided file. 
	 * If no factory can be found to handle the file then null is returned. 
	 * <p>
	 * File names without extension will always return null  
	 *  
	 * @throws IllegalArgumentException if file is null
	 * @throws IOException if errors occur during image creation
	 * 
	 * @param file to create image from
	 * @return Image or null
	 * 
	 */
	public Image getImage(File file) throws IOException {
			if(file==null) throw new IllegalArgumentException("file is null");
		
			String fileName=file.getName();
			
			int lastDot=fileName.lastIndexOf('.');
			
			if(lastDot<0) {  
				return null; // no suffix
			}
			
			if( (lastDot+1)==fileName.length()) {
				return null;  // no suffix
			}
			
			String suffix=fileName.substring(lastDot+1);
			
			for(ImageFactory factory:factories) {
				String[] extensions=factory.getValidFileExtensions();
				if(extensions==null) continue;
				for(String value:extensions) {
					if(suffix.equalsIgnoreCase(value)) {
						// found match
						return factory.getImage(file);
					}
				}
				
			}
			return null; // no matches
	}

	/**
	 * <p>
	 * Returns an appropriate {@link javax.tools.diagnostics.image.Image} for the
	 * provide file by locating the first registered image factory that can handle
	 * the case insensitive file name extension of the provided file. 
	 * If no factory can be found to handle the file then null is returned. 
	 * <p>
	 * File names without extension will always return null  
	 *  
	 * @throws IllegalArgumentException if file is null
	 * @throws IOException if errors occur during image creation
	 * 
	 * @param file to create image from
	 * @return Image or null
	 * 
	 */
	public JavaRuntime getJavaRuntime(File file) throws IOException {
			if(file==null) throw new IllegalArgumentException("file is null");
		
			String fileName=file.getName();
			
			int lastDot=fileName.lastIndexOf('.');
			
			if(lastDot<0) {  
				return null; // no suffix
			}
			
			if( (lastDot+1)==fileName.length()) {
				return null;  // no suffix
			}
			
			String suffix=fileName.substring(lastDot+1);
			
			for(ImageFactory factory:factories) {
				String[] extensions=factory.getValidFileExtensions();
				if(extensions==null) continue;
				for(String value:extensions) {
					if(suffix.equalsIgnoreCase(value)) {
						// found match
						return factory.getJavaRuntime(file);
						
					}
				}
				
			}
			return null; // no matches
	}
}
