/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package javax.tools.diagnostics.image;

import java.io.File;
import java.net.InetAddress;
import java.util.Iterator;
import java.util.List;

/**
 * This class represents an entire operating system image (e.g. a core file). 
 * 
 * There are methods for accessing information about the architecture 
 * of the machine on which the image was running - hardware and 
 * operating system. The major feature, however, is the ability to 
 * iterate over the Address Spaces contained within the image. 
 *
 */
public interface Image {
	
	/**
	 * Get the set of address spaces within the image - typically one but may be more on some 
	 * systems such as z/OS. 
	 * @return an Iterator which iterates over all of the address spaces
	 * described by this Image
	 * 
	 * @see ImageAddressSpace
	 */
	List<ImageAddressSpace> getAddressSpaces();

	/**
	 * Get the family name for the processor on which the image was
	 * running.
	 * @return the family name for the processor on which the image was
	 * running. This corresponds to the value you would find in the
	 * "os.arch" System property.
	 * 
	 * @throws DataUnavailable if this data cannot be inferred from this core type
	 * @throws CorruptDataException if expected data cannot be read from the core
	 */
	String getProcessorType() throws DataUnavailable, CorruptDataException;
	
	/**
	 * Get the precise model of the CPU.
	 * @return the precise model of the CPU (note that this can be an empty string but not null).
	 * <br>
	 * e.g. getProcessorType() will return "x86" where getProcessorSubType() may return "Pentium IV step 4"
	 * <p>
	 * Note that this value is platform and implementation dependent.
	 * @throws DataUnavailable 
	 * @throws CorruptDataException 
	 */
	String getProcessorSubType() throws DataUnavailable, CorruptDataException;
	
	/**
	 * Get the number of CPUs running in the system on which the image was running.
	 * @return the number of CPUs running in the system on which the
	 * image was running
	 * 
	 * @exception DataUnavailable if the information cannot be provided
	 */
	int getProcessorCount() throws DataUnavailable;
	
	/**
	 * Get the family name for the operating system.
	 * @return the family name for the operating system.  This should be the same value
	 * that would be returned for the "os.name" system property
	 * 
	 * @throws DataUnavailable if this data cannot be inferred from this core type
	 * @throws CorruptDataException if expected data cannot be read from the core
	 */
	String getSystemType() throws DataUnavailable, CorruptDataException;
	
	/**
	 * Get the detailed name of the operating system.
	 * @return the detailed name of the operating system, or an empty string
	 * if this information is not available (null will never be returned).  This should be
	 * the same value that would be returned for the "os.version" system property
	 * @throws DataUnavailable 
	 * @throws CorruptDataException 
	 */
	String getSystemSubType() throws DataUnavailable, CorruptDataException;
	
	/**
	 * Get the amount of physical memory (in bytes) installed in the system on which
	 * the image was running.
	 * @return the amount of physical memory installed in the system on which
	 * the image was running.  The return value is specified in bytes.
	 * 
	 * @exception DataUnavailable if the information cannot be provided
	 */
	long getInstalledMemory() throws DataUnavailable;
	
	/**
	 * Determines when the image was created
	 * 
	 * @return the time in milliseconds since 1970
	 * @throws DataUnavailable 
	 */
	long getCreationTime() throws DataUnavailable;
	
	/**
	 * Get the host name of the system where the image was running. 
	 * @return The host name of the system where the image was running.  This string will
	 * be non-null and non-empty
	 * 
	 * @throws DataUnavailable If the image did not provide this information (would happen
	 * if the system did not know its host name or if the image predated this feature).
	 * @throws CorruptDataException 
	 * 
	 * @since 1.0
	 */
	public String getHostName() throws DataUnavailable, CorruptDataException;
	
	/**
	 * The set of IP addresses (as InetAddresses) which the system running the image possessed. 
	 * @return An Iterator over the IP addresses (as InetAddresses) which the system running 
	 * the image possessed.  The iterator will be non-null (but can be empty if the host is 
	 * known to have no IP addresses).
	 * 
	 * @throws DataUnavailable If the image did not provide this information (would happen
	 * if the system failed to look them up or if the image pre-dated this feature).
	 * 
	 * @since 1.0
	 * 
	 * @see java.net.InetAddress
	 */
	public Iterator<InetAddress> getIPAddresses() throws DataUnavailable;
	
	/**
	 * Experimenal
	 */
	public File getSource();
}
