/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package javax.tools.diagnostics.image;

import java.io.File;
import java.io.IOException;

import javax.tools.diagnostics.runtime.java.JavaRuntime;

/**
 * <p>
 * This interface is used for classes which can produce instances of Image
 * implementors.
 * <p>
 * Note that this interface forms the contract between the {@link javax.tools.diagnostics.FactoryRegistry}
 * and an implementation.   The methods on this interface are only intended to be called by the  {@link javax.tools.diagnostics.FactoryRegistry}
 *  
 */

public interface ImageFactory {
	
	/**
	 * <p>
	 * The major version number of the current API.
	 * <p>
	 * Note that this value will be inlined so users can reference it directly
	 * (ie:  factory.getMajorVersion() >= ImageFactory.MAJOR_VERSION)
	 */
	public static final int MAJOR_VERSION = 0;
	
	/**
	 * <p>
	 * The minor version number of the current API
	 * <p?
	 * Note that this value will be inlined so users can reference it directly
	 * (ie:  factory.getKatoMinorVersion() >= ImageFactory.MINOR_VERSION)
	 */
	public static final int MINOR_VERSION = 1;
	
	/**
	 * <p>
	 * Creates a new Image object based on the contents of imageFile.
	 * <p>
	 * Note that this method is only intended to be called by the {@link javax.tools.diagnostics.FactoryRegistry}
	 * 
	 * @param imageFile a file with Image information, typically a core file
	 * @return an instance of Image
	 * @throws IOException if unable to create an image from the provided file 
	 */
	Image getImage(File imageFile) throws IOException;

	/**
	 * <p>
	 * Creates a new Image object based on the contents of imageFile and metadata
	 * <p>
	 * Note that this method is only intended to be called by the {@link javax.tools.diagnostics.FactoryRegistry}
	 * @param imageFile a file with Image information, typically a core file
	 * @param metadata a file with additional Image information. This is an implementation defined file
	 * @return an instance of Image
	 * @throws IOException if unable to create an image from the provided file
	 */
	Image getImage(File imageFile, File metadata) throws IOException;
	
	/**
	 * <p>
	 * Fetch the major version number
	 * @return An integer corresponding to the API major version number
	 */
	public int getMajorVersion();
	
	/**
	 * <p>
	 * Fetch the minor version number
	 * @return An integer corresponding to the API minor version number
	 */
	public int getMinorVersion();
	
	/**
	 * <p>
	 * Fetch the modification level
	 * @return An integer corresponding to the  API modification level
	 * 
	 */	
	public int getModificationLevel();

	/**
	 * <p>
	 * Returns an array of file extensions that the {@link javax.tools.diagnostics.FactoryRegistry} can use
	 * to determine if an file can be processed by this Image Factory implementation.
	 * <p>File extensions are the part of a file name after the last '.'. 
	 * <p>
	 * The returned array is treated as a case insensitive collection of file extensions.
	 * <p>
	 * The returned array is expected to contain at least one entry.
 	 * <p>
	 * Note that this method is only intended to be called by the {@link javax.tools.diagnostics.FactoryRegistry}

	 * @return an array of case insensitive file extension names.
	 */
	public String[] getValidFileExtensions();
	
	/**
	 * <p>
	 * Creates a new JavaRuntime object based on the contents of the file;
	 * <p>
	 * Note that this method is only intended to be called by the {@link javax.tools.diagnostics.FactoryRegistry}
	 * 
	 * @param file a file with JavaRuntime information
	 * @return an instance of JavaRuntime
	 * @throws IOException if unable to create a runtime from the provided file 
	 */
	public JavaRuntime getJavaRuntime(File file) throws IOException;
}
