/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package javax.tools.diagnostics.image;

import java.util.List;
import java.util.Properties;

/**
 * Represents a shared library loaded into the image, or the executable module itself
 * 
 */
public interface ImageModule {

    /**
     * Get the file name of the shared library.
     * @return the file name of the shared library
     * 
     * @throws CorruptDataException If the module is corrupt and the 
     * original file cannot be determined
     */
    String getName() throws CorruptDataException;
    
    /**
     * Get the collection of sections that make up this library.
     * @return a collection of sections that make up this library
     * 
     * @see ImageSection
     */
    List<ImageSection> getSections();
    
    /**
     * Provides a collection of symbols defined by the library. This
     * list is likely incomplete as many symbols may be private,
     * symbols may have been stripped from the library, or symbols may
     * not by available in the image.
     * 
     * @return a collection of symbols which are defined by this library.
     * 
     * @see ImageSymbol
     */
    List<ImageSymbol> getSymbols();
    
    /**
     * Get the table of properties associated with this module.
     * @return a table of properties associated with this module.
     * Values typically defined in this table include
     * <ul>
     * <li>"version" -- version information about the module</li>
     * </ul>
     * @throws CorruptDataException 
     */
    Properties getProperties() throws CorruptDataException;
    
}
