/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package javax.tools.diagnostics.runtime.java;

import javax.tools.diagnostics.image.CorruptDataException;
import javax.tools.diagnostics.image.MemoryAccessException;

/**
 * <p>
 * Represents a field declaration.
 * It is modelled on {@link java.lang.reflect.Field}
 */
public interface JavaField extends JavaMember {

    /**
     * <p>
     * Get the contents of a field of an Object.
     * 
     * @param object to fetch the field from. Ignored for static
     * fields.
     * 
     * 
     * <p>
     * This field must be declared in the object's class or in a superclass
     * <p>
     * @return a JavaObject instance for reference type fields,
     * an instance of a subclass of Number, Boolean, or Character 
     * for primitive fields, or null for null reference fields.
     * 
     * @throws CorruptDataException if the underlying data is in an unexpected state 
     * @throws MemoryAccessException  
     * @throws NullPointerException if the field is an instance field, and object is null
     * @throws IllegalArgumentException if the specified object is not appropriate for
     * this field
     *
     * @see JavaObject
     * @see Byte
     * @see Double
     * @see Float
     * @see Integer
     * @see Long
     * @see Short
     * @see Character
     * @see Boolean
     */
    Object get(JavaObject object) throws CorruptDataException, MemoryAccessException;

    /**
     * <p>
     * Get the contents of a boolean field
     * @param  object to fetch the field from. Ignored for static fields.
     * @return the field contents
     * @throws CorruptDataException if the underlying data is in an unexpected state 
     * @throws MemoryAccessException 
     * @throws NullPointerException if the field is an instance field, and object is null
     * @throws IllegalArgumentException if the specified object is not appropriate for
     * this field, or if the field is not a boolean.
     */
    boolean getBoolean(JavaObject object) throws CorruptDataException, MemoryAccessException;

    /**
     * <p>
     * Get the contents of a byte field
     * @param  object to fetch the field from. Ignored for static fields.
     * @return the field contents
     * @throws CorruptDataException if the underlying data is in an unexpected state 
     * @throws MemoryAccessException 
     * @throws NullPointerException if the field is an instance field, and object is null
     * @throws IllegalArgumentException if the specified object is not appropriate for
     * this field, or if the type of the field cannot be converted to byte
     */
    byte getByte(JavaObject object) throws CorruptDataException, MemoryAccessException;

    /**
     * <p>
     * Get the contents of a char field
     * @param  object to fetch the field from. Ignored for static fields.
     * @return the field contents
     * @throws CorruptDataException if the underlying data is in an unexpected state 
     * @throws MemoryAccessException 
     * @throws NullPointerException if the field is an instance field, and object is null
     * @throws IllegalArgumentException if the specified object is not appropriate for
     * this field, or if the type of the field cannot be converted to char
     */
    char getChar(JavaObject object) throws CorruptDataException, MemoryAccessException;

    /**
     * <p>
     * Get the contents of a double field or of another primitive field whose type is convertible to double via a widening conversion. 
     * @param  object to fetch the field from. Ignored for static fields.
     * @return the field contents
     * @throws CorruptDataException if the underlying data is in an unexpected state 
     * @throws MemoryAccessException 
     * @throws NullPointerException if the field is an instance field, and object is null
     * @throws IllegalArgumentException if the specified object is not appropriate for
     * this field, or if the type of the field cannot be converted to double via a widening conversion
     */
    double getDouble(JavaObject object) throws CorruptDataException, MemoryAccessException;

    /**
     * <p>
     * Get the contents of a float field or of another primitive field whose type is convertible to float via a widening conversion. 
     * @param  object to fetch the field from. Ignored for static fields.
     * @return the field contents 
     * @throws CorruptDataException if the underlying data is in an unexpected state 
     * @throws MemoryAccessException 
     * @throws NullPointerException if the field is an instance field, and object is null
     * @throws IllegalArgumentException if the specified object is not appropriate for
     * this field, or if the type of the field cannot be converted to float via a widening conversion
     */
    float getFloat(JavaObject object) throws CorruptDataException, MemoryAccessException;

    /**
     * <p>
     * Get the contents of an int field or of another primitive field whose type is convertible to int via a widening conversion.
     * @return the field contents
     * @param  object to fetch the field from. Ignored for static fields.
     * @throws CorruptDataException if the underlying data is in an unexpected state 
     * @throws MemoryAccessException 
     * @throws NullPointerException if the field is an instance field, and object is null
     * @throws IllegalArgumentException if the specified object is not appropriate for
     * this field, or if the type of the field cannot be converted to int via a widening conversion.
     */
    int getInt(JavaObject object) throws CorruptDataException, MemoryAccessException;

    /**
     * <p>
     * Get the contents of a long field or of another primitive field whose type is convertible to long via a widening conversion.
     * @param  object to fetch the field from. Ignored for static fields.
     * @return the field contents
     * @throws CorruptDataException if the underlying data is in an unexpected state 
     * @throws MemoryAccessException 
     * @throws NullPointerException if the field is an instance field, and object is null
     * @throws IllegalArgumentException if the specified object is not appropriate for
     * this field, or if the type of the field cannot be converted to long via a widening conversion.
     */
    long getLong(JavaObject object) throws CorruptDataException, MemoryAccessException;

    /**
     * <p>
     * Get the contents of a short field or of another primitive field whose type is convertible to short via a widening conversion.
     * @param  object to fetch the field from. Ignored for static fields.
     * @return the field contents 
     * @throws CorruptDataException if the underlying data is in an unexpected state 
     * @throws MemoryAccessException 
     * @throws NullPointerException if the field is an instance field, and object is null
     * @throws IllegalArgumentException if the specified object is not appropriate for
     * this field, or if the type of the field cannot be converted to short via a widening conversion.
     */
    short getShort(JavaObject object) throws CorruptDataException, MemoryAccessException;
    
    /**
     * <p>
     * Get the contents of a string field
     * @param object to fetch the field from. Ignored for static fields.
     * 
     * @return a String representing the value of the String field.  Note that the instance
     * returned can be null if the field was null in object.
     * @throws CorruptDataException if the underlying data is in an unexpected state 
     * @throws MemoryAccessException 
     * 
     * @throws IllegalArgumentException if the specified field is not a String
     * @throws NullPointerException if the field is an instance field, and object is null
     */
    String getString(JavaObject object) throws CorruptDataException, MemoryAccessException;
    
	/**
	 * @param obj
	 * @return True if the given object refers to the same Java Field in the image
	 */
	public boolean equals(Object obj);
	public int hashCode();
}
