/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package javax.tools.diagnostics.runtime.java;

import javax.tools.diagnostics.image.CorruptDataException;
import javax.tools.diagnostics.image.DataUnavailable;

/**
 * <p>
 * Abstract interface which both JavaField and JavaMethod inherit from.
 * It defines APIs which are common to both types of members.
 * It is modelled on java.lang.reflect.Member
 */
public interface JavaMember {

	/**
	 * <p>
	 * Get the set of modifiers for this field or method - a set of bits
	 * The values for the constants representing the modifiers can be obtained from {@link java.lang.reflect.Modifier}.

	 * @return the modifiers for this field or method. 
	 * @throws CorruptDataException if the underlying data is in an unexpected state  
     *
     */
    int getModifiers()  throws CorruptDataException;

    /**
     * <p>
     * Get the class which declares this field or method
     * 
     * @return the JavaClass which declared this field or method
     * @throws CorruptDataException if the underlying data is in an unexpected state  
     * @throws DataUnavailable if there is no declaring class available
     */
    JavaClass getDeclaringClass() throws CorruptDataException, DataUnavailable;
    
    /**
     * <p>
     * Get the name of the field or method
     * @return the name of the field or method
     * @throws CorruptDataException if the underlying data is in an unexpected state  
     */
    String getName()  throws CorruptDataException;
    
    /**
     * <p>
     * Get the signature of the field or method
     * @return the signature of the field or method.
     * e.g. "(Ljava/lang/String;)V"
     * @throws CorruptDataException if the underlying data is in an unexpected state  
     */
    String getSignature()  throws CorruptDataException;
    
	/**
	 * @param obj
	 * @return True if the given object refers to the same Java Member in the image
	 */
	public boolean equals(Object obj);
	public int hashCode();
}
