/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package javax.tools.diagnostics.runtime.java;


import java.util.List;

import javax.tools.diagnostics.image.CorruptDataException;
import javax.tools.diagnostics.image.DataUnavailable;
import javax.tools.diagnostics.image.ImagePointer;
import javax.tools.diagnostics.image.ImageSection;
import javax.tools.diagnostics.image.MemoryAccessException;



/**
 * <p>Represents a Java object or array.</p>
 * <p>Array elements can be retrieved using the {@link #arraycopy()} method. Object instance fields
 * can be retrieved using the <code>get*()</code> methods in {@link JavaField} such as
 * {@link JavaField#get(JavaObject)}. The {@link JavaField} objects can be retrieved from an
 * object's {@link JavaClass} using {@link JavaClass#getDeclaredFields()}. 
 * </p> 
 * 
 * 
 */
public interface JavaObject {
    
    /**
     * <p>Get the JavaClass instance which represents the class of this object.</p>
     * <p>This method never returns null, all objects have a class. The JavaClass returned
     * might be synthetic for array types.</p>
     * 
     * @return the JavaClass instance which represents the class of this object.
     * @throws CorruptDataException 
     */
    JavaClass getJavaClass() throws CorruptDataException;
    
    /**
     * <p>Returns true if this JavaObject represents an array.</p>
     * 
     * @return true if this JavaObject represents an array.
     * @throws CorruptDataException 
     */
    boolean isArray()  throws CorruptDataException;
    
    /**
     * <p>Get the number of elements in this array.</p>
     * <p>This is equivalent to calling <code>array.length</code> in Java, where <code>array</code> is an
     * array reference.</p>
     * 
     * @return the number of elements in this array.
     * @throws CorruptDataException 
     * @exception IllegalArgumentException if the object is not an array.
     */
    int getArraySize() throws CorruptDataException;
    
    /**
     * <p>Copies data from the array this JavaObject represents into an array.</p>
     * <p>The dst object must be an array of the appropriate type --
     * a primitive type array for base types, or a JavaObject array
     * for reference arrays.</p> 
     *  
     * @param srcStart index in the receiver to start copying from.
     * @param dst the destination array.
     * @param dstStart index in the destination array to start copying into.
     * @param length the number of elements to be copied.
     * @throws CorruptDataException 
     * @throws MemoryAccessException 
     * 
     * @exception NullPointerException if dst is null.
     * @exception IllegalArgumentException if the object is not an array,
     * or if dst is not an array of the appropriate type.
     * @exception IndexOutOfBoundsException if srcStart, dstStart, or length
     * are out of bounds in either the JavaObject or the destination array.
     */
    void arraycopy(int srcStart, Object dst, int dstStart, int length)  throws CorruptDataException, MemoryAccessException;
    
    /**
     * <p>Get the number of bytes of memory occupied by this object.</p>
     * 
     * @return the number of bytes of memory occupied by this object. The
     * memory may not necessarily be contiguous.
     * @throws CorruptDataException 
     */
    long getSize()  throws CorruptDataException;
    
    /**
     * <p>Fetch the basic hash code for the object.</p>
     * <p> This is the hash code which would be returned if a Java thread had 
     * requested it. Typically the hash code is based on the address of an 
     * object, and may change if the object is moved by a garbage collect cycle.
     * 
     * @return the basic hash code of the object in the image.
     * 
     * @see #getPersistentHashcode()
     * 
     * @exception DataUnavailable if the hash code cannot be determined.
     * @throws CorruptDataException 
     */
    long getHashcode() throws DataUnavailable, CorruptDataException;
    
    /**
     * Fetch the basic hash code of the object in the image. This hash code
     * is guaranteed to be persistent between multiple snapshots of the same Image.
     * If the hash code cannot be determined, or if the hash code for this object
     * could change between snapshots, an exception is thrown.
     * <p>
     * <i>If the VM uses a 'hasBeenHashed' bit, the value of this bit can be
     * inferred by calling getPersistentHashcode(). If the persistent hash code
     * is not available, then the 'hasBeenHashed' bit has not been set, and the
     * hash of the object could change if the object moves between snapshots</i>
     * 
     * @return the basic hash code of the object in the image
     * 
     * @see #getHashcode()
     * 
     * @exception DataUnavailable if a hash code cannot be determined, or if the
     * hash code could change between successive snapshots
     * @throws CorruptDataException 
     */
    long getPersistentHashcode() throws DataUnavailable, CorruptDataException;
    
    /**
     * <p>The ID of an object is a unique address is memory which identifies the object.
     * </p>
     * <p>It is probable that an object's address will change during the lifetime of a Java Virtual
     * Machine because of the operations of the garbage collector. Other mechanisms for uniquely identifying
     * objects should be used when comparing dumps.
     * </p>
     * <p>
     * The data at this memory is implementation defined. The object may be 
     * non-contiguous. Portions of the object may appear below or above this address.</p>
     * </p>
     * 
     * @return the runtime-wide unique identifier for the object.
     * 
     */
	ImagePointer getID();
	
    /**
     * <p>Returns the sections that this object occupies in memory.</p>
     * <p>These sections include the object's header and the data in the object.</p>
     *
     * <p>In certain allocation strategies, an object's header and data may be allocated 
     * contiguously. In this case, this method may return an iterator for a single
     * section.</p>
     * 
     * <p>In other schemes, the header may be separate from the data or the data may be
     * broken up into multiple regions. Additionally, this function does not guarantee
     * that the memory used by this object is not also shared by one or more other
     * objects.</p>
     * 
     * <p>The contents of the image sections are implementation specific, as so are
     * undefined here.</p> 
     * 
     * @return a collection of sections that make up this object.
     * 
     * @see ImageSection
     */
    List<ImageSection> getSections();
    
	/**
	 * <p>Get the set of references from this object.</p>
	 * <p>These references will include at least the object's references to its class,
	 * and any references from instance fields to other objects and classes, or array elements
	 * references to other objects.</p>
	 *
	 * @return an List of JavaReferences.
	 * @see javax.tools.diagnostics.runtime.java.JavaReference
	 *
	 */
	List<JavaReference> getReferences();
    
    /**
     * <p>Gets the heap where this object is located.</p>
     * <p>A JavaHeap will always be returned if this object could be retrieved by
     * {@link JavaHeap#getObject()}, otherwise {@link DataUnavailable} is thrown.
     * </p>
     * 
     * @return	the {@link JavaHeap} instance representing the heap where this object is stored in memory.
     * @throws 	CorruptDataException	if the heap information for this object is corrupt.
     * @throws 	DataUnavailable			if the heap information for this object is not available.
     * @see 	JavaHeap
     */
    JavaHeap getHeap() throws CorruptDataException, DataUnavailable;
	
	/**
	 * @param obj
	 * @return True if the given object refers to the same Java Object in the image
	 */
	public boolean equals(Object obj);
	public int hashCode();
}
