/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/

package org.apache.kato.jvmti.javaruntime.model;

import java.util.Collections;
import java.util.LinkedList;
import java.util.List;

import javax.tools.diagnostics.image.CorruptDataException;
import javax.tools.diagnostics.image.ImagePointer;
import javax.tools.diagnostics.runtime.java.JavaClass;
import javax.tools.diagnostics.runtime.java.JavaClassLoader;
import javax.tools.diagnostics.runtime.java.JavaMethod;
import javax.tools.diagnostics.runtime.java.JavaObject;
import javax.tools.diagnostics.runtime.java.JavaReference;

import org.apache.kato.jvmti.process.CorruptDataImpl;

public class JClass implements JavaClass {
	/**
	 * 
	 */
	private Model model = null;

	/**
	 * @param binReader
	 */
	JClass(Model model) {
		this.model = model;
	}

	JClass() {

	}

	public long classid = 0;
	public long superClassID;
	public JClass superclass = null;
	public long classloaderID;
	public JClassLoader classloader = null;
	public short status = 0;
	public short modifiers = 0;
	public short fieldCount = 0;
	public short interfaceCount = 0;
	public short methodsCount = 0;
	public int nameLength = 0;
	public String classSig = null;
	public String sourceFile;
	private List<JClass> interfaces = null;
	private List<JavaMethod> methods = null;
	private ObjectMapList<Long, JField> fields = null;
	public int locIn; // remove this

	@Override
	public JavaClassLoader getClassLoader() throws CorruptDataException {
		return classloader;
	}

	public void addInterface(JClass face) {
		if (interfaces == null) {
			interfaces = new LinkedList<JClass>();
		}
		interfaces.add(face);

	}

	public void addMethod(JMethod m) {
		if (methods == null) {
			methods = new LinkedList<JavaMethod>();
		}
		m.setParent(this);
		methods.add(m);
	}

	private JClass componentType = null;
	
	/**
	 * Returns the component type of this array class. 
	 * Takes the 
	 * 
	 * @return JavaClass
	 * @throws IllegalArgumentException when the JavaClass is not an array type/
	 */
	@Override
	public JavaClass getComponentType() throws CorruptDataException {
		if (!isArray()) {
			throw new IllegalArgumentException("JavaClass.getComponentType() cannot be called on a non-array type - \"" +
					classSig + "\"");
		}
		
		
		if (componentType == null) {
			String componentSig = classSig.substring(1);
					
			switch (componentSig.charAt(0)) {
			case 'L':
				// This is an array of objects. Strip off L and ;
				// [Ljava/lang/Object; -> java/lang/Object
				componentSig = componentSig.substring(1, componentSig.length()-1);				
				break;
			case '[':
				// This is fine - make no change. 
				break;
			case 'Z':
				componentSig = "boolean";
				break;
			case 'B':
				componentSig = "byte";
				break;
			case 'C':
				componentSig = "char";
				break;
			case 'S':
				componentSig = "short";
				break;
			case 'I':
				componentSig = "int";
				break;
			case 'J':
				componentSig = "long";
				break;
			case 'F':
				componentSig = "float";
				break;
			case 'D':
				componentSig = "double";
				break;
			default:
				throw CorruptDataImpl.corruptDataException("Tried to determine component type from invalid starting character '" +
						+componentSig.charAt(0)+"', should be one of 'ZBCSIJFD[L'");
			}
			componentType = (JClass) classloader.findClass(componentSig);
			
			if (componentType == null) {
				JavaObject classLoaderObj =classloader.getObject();
				long id = 0;
				
				if (classLoaderObj != null) {
					id = classLoaderObj.getID().getAddress();
				}
				
				throw CorruptDataImpl.corruptDataException("Unable to find class of type `"+componentSig+"' from" +
						" classloader 0x"+Long.toHexString(id));
			}
		}
		return componentType;
	}

	@Override
	public List getConstantPoolReferences() {
		return Collections.EMPTY_LIST;

	}

	@Override
	public List getDeclaredFields() {

		if (fields == null)
			fields = new ObjectMapList<Long, JField>();
		return Collections.unmodifiableList(fields.values());
	}

	@Override
	public List<JavaMethod> getDeclaredMethods() {
		if (methods == null)
			methods = new LinkedList<JavaMethod>();
		return Collections.unmodifiableList(methods);
	}

	@Override
	public ImagePointer getID() {
		return new SimpleImagePointer(classid);
	}

	@Override
	public List<String> getInterfaces() {
		// TODO implement this
		LinkedList<String> interfaceNames = new LinkedList<String>();
		if (interfaces != null) {
			for (JClass interfaceClass : interfaces) {
				try {
					interfaceNames.add(interfaceClass.getName());
				} catch (CorruptDataException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			}
		}
		return interfaceNames;
	}

	@Override
	public int getModifiers() throws CorruptDataException {
		return modifiers;
	}

	@Override
	public String getName() throws CorruptDataException {
		if (!isArray()) {
			return classSig.substring(1, classSig.length() - 1);
		}
		return classSig;
	}

	@Override
	public JavaObject getObject() throws CorruptDataException {
		// This is expected.
		return null;
	}

	@Override
	public List<JavaReference> getReferences() {
		return Collections.emptyList();
	}

	@Override
	public JavaClass getSuperclass() throws CorruptDataException {
		return superclass;
	}

	@Override
	public boolean isArray() throws CorruptDataException {
		return classSig.startsWith("[");
	}

	public JField getField(long id) {

		if (fields == null) {
			fields = new ObjectMapList<Long, JField>();
		}

		JField f = (JField) fields.get(id);
		if (f == null) {
			f = new JField(this, id);
			fields.put(id, f);
		}
		return f;

	}
}

