/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.apache.kato.hprof.java;


import java.util.LinkedList;
import java.util.List;

import javax.tools.diagnostics.image.CorruptDataException;
import javax.tools.diagnostics.image.DataUnavailable;
import javax.tools.diagnostics.image.ImagePointer;
import javax.tools.diagnostics.image.MemoryAccessException;
import javax.tools.diagnostics.runtime.java.JavaClass;
import javax.tools.diagnostics.runtime.java.JavaHeap;
import javax.tools.diagnostics.runtime.java.JavaObject;

import org.apache.kato.hprof.datalayer.IGCObjectArrayHeapDumpRecord;

/**
 * An array of object references.
 * 
 * 
 *
 */
public class JavaObjectArrayImpl implements JavaObject {
	private IJavaHeapInternal heap;
	private IGCObjectArrayHeapDumpRecord record;

	public JavaObjectArrayImpl(IJavaHeapInternal heap, IGCObjectArrayHeapDumpRecord record) {
		this.heap = heap;
		this.record = record;		
	}

	/**
	 * Retrieves object IDs from the record on the disk and fills in array
	 * with JavaObjects by mapping from the IDs to JavaObjects.
	 * 
	 */
	@Override
	public void arraycopy(int srcStart, Object dst, int dstStart, int length)
			throws CorruptDataException, MemoryAccessException {
		if (srcStart < 0 || srcStart > getArraySize()) {
			throw new IndexOutOfBoundsException("srcStart must be between 0 and " + getArraySize()+", not " + srcStart);
		}
		
		if (srcStart+length > getArraySize()) {
			throw new IndexOutOfBoundsException("srcStart+length must be <="+getArraySize()+", srcStart=" + srcStart+", length="+length);			
		}
		if (dstStart < 0) {
			throw new IndexOutOfBoundsException("dstStart must be >=0, not " + dstStart);
		}

		if (dst == null) {
			throw new NullPointerException("Destination array must not be null");
		}
		if (!dst.getClass().isArray()) {
			throw new IllegalArgumentException("Destination array must be an array, not "+dst.getClass().getName());			
		}
		if(dst.getClass().getComponentType() != JavaObject.class) {
			throw new IllegalArgumentException("Destination array wrong type - must be JavaObject array");
		}
		
		JavaObject[] array = (JavaObject[]) dst;
		checkArrayBounds(dstStart, array.length, length);
		
		for (int i=0; i < length; i++) {
			array[dstStart+i] = heap.getObjectByID(record.getElement(srcStart+i));
		}
	}
	
	private static void checkArrayBounds(int arrayStart, int arrayLength, int dataLength) {
		if ( arrayStart + dataLength > arrayLength) {
			throw new IndexOutOfBoundsException("Destination array of length "+ arrayLength+" can't" 
					+ " accomodate array copy with start of "+arrayStart+" and length of "+ dataLength);
		}
	}
	@Override
	public int getArraySize() throws CorruptDataException {
		return record.getNumberOfElements();
	}

	@Override
	public long getHashcode() throws DataUnavailable, CorruptDataException {
		// this is probably a bad hash
		return record.getID(); 
	}

	@Override
	public JavaHeap getHeap() throws CorruptDataException, DataUnavailable {
		return heap;
	}

	@Override
	public ImagePointer getID() {
		return heap.getImagePointer(record.getID());
	}

	@Override
	public JavaClass getJavaClass() throws CorruptDataException {
		return heap.getJavaClassByID(record.getArrayClassObjectID());
	}

	@Override
	public long getPersistentHashcode() throws DataUnavailable,
			CorruptDataException {
		throw new DataUnavailable("Persistent hashcode unknown.");	}

	@Override
	public List getReferences() {
		return new LinkedList();
	}

	@Override
	public List getSections() {
		return new LinkedList();
	}

	@Override
	public long getSize() throws CorruptDataException {
		return 0;
	}

	@Override
	public boolean isArray() throws CorruptDataException {
		return true;
	}
	
	/**
	 * Compares identity. This object may be created multiple times.
	 * Uses object ID for comparing identity. 
	 * We don't check runtime identity.
	 */
	@Override
	public boolean equals(Object obj) {
		if (obj == null) {
			return false;
		}
		
		if (!(obj instanceof JavaObjectArrayImpl)) {
			JavaObjectArrayImpl other = (JavaObjectArrayImpl) obj;
			
			if (other.getObjectID() == record.getID()) {
				return true;
			}
		}
		
		return false;
	}
	
	/**
	 * Return object ID of record on disk.
	 * @return
	 */
	public long getObjectID() {
		return record.getID();
	}
}
