/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.apache.kato.tools.jdi;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.PrintStream;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.Arrays;
import java.util.Vector;

/* JDICacher
 * 
 * JDICacher lies as a layer between the JDWPServer and the KatoReader.  It times the returns
 * from KatoReader, and if they're over a certain amount of time, it stores them in a vector
 * so that future calls return the vectorized version.  The actual class that's stored in the
 * Vector is a CommandReplyPair, which stores a command and it's reply.  On read, we search
 * all the pairs for a matching CommandPacket, and directly return it's ReplyPacket.  Most of
 * the time this isn't an issue, but sometimes it can be quite lengthy to do some of the
 * operations on larger core files.
 * 
 *   It should be noted that we always do an MD5 hash on the core file on our first cache request
 * (if and only if file caching is enabled).  This may be a large bottleneck any of:
 *  o No hashing
 *  o No file cache
 *  o No cache
 * will alleviate this problem.  It should be noted to use 'no hashing' with extreme care, because
 * if you load another cache from another file, you'll start serving out IDs which are invalid for
 * this core.
 * 
 *   Also, delay cache write is recommended for over-the-network cache saves, because by default we'll
 * re-write the cache to disk on every new packet added.  Most of the time this is what we want, but
 * if you don't properly shut down the server (the red square in the debugger), no cache will actually
 * be saved.
 */

public class JDICacher{
	private String coreFile;

	//This caches everything set to cache
	public static final int CACHE_CACHE = 0;
	//Cache only things set to cache
	public static final int CACHE_EXPLICIT = 1;
	//Do not cache anything
	public static final int CACHE_NOCACHE = 2;
	//Cache everything
	public static final int CACHE_FORCECACHE = 3;
	private int cacheTime = 1000;

	private JDILogger logr;
	private int cacheSettings = CACHE_CACHE;
	private CacherProvider provider;
	Vector<CommandReplyPair> cache;
	private boolean disableHash = false;
	private boolean doDelay = false;
	public final static int FILE_NOFILE = 0;
	public final static int FILE_USEFILE = 1;
	private int fileSettings = FILE_USEFILE;
	private String hashCode;
	
	public JDICacher(String coreFile, JDWPServer svr, JDILogger logger) throws Exception{
		logr = logger;
		logr.log(JDILogger.LEVEL_VERYVERBOSE, "JDICacher 1.0"); //$NON-NLS-1$
		this.coreFile = coreFile;
		logr.log(JDILogger.LEVEL_VERYVERBOSE, "Creating KatoReader"); //$NON-NLS-1$
		provider = new KatoReader(coreFile, svr, JDILogger.logger);
		logr.log(JDILogger.LEVEL_VERYVERBOSE, "KatoReader created"); //$NON-NLS-1$




	}

	public void setProvider(CacherProvider pvdr){
		provider = pvdr;
	}

	public void setCacheSettings(int cacheSettings){
		this.cacheSettings = cacheSettings;
	}

	public int getCacheSettings(){
		return this.cacheSettings;
	}

	public void commit(){
		if (fileSettings == FILE_USEFILE){
			resaveCache();
		}
	}

	public void setFileAccess(int setting){
		fileSettings = setting;
	}

	public void delayedWrites(boolean inDoDelay){
		this.doDelay = inDoDelay;
	}

	public void disableHash(boolean noHash){
		this.disableHash = noHash;
	}

	private ReplyPacket readCache(CommandPacket cmdpckt){

		//If we need initing, init
		if (this.hashCode == null){
			this.hashCode = getHash();
		}
		if (cache == null){
			if (fileSettings == FILE_USEFILE){
				try{
					FileInputStream bStream = new FileInputStream(hashCode + ".dat"); //$NON-NLS-1$
					ObjectInputStream oStream = new ObjectInputStream(bStream);
					logr.log(JDILogger.LEVEL_VERBOSE, "Reading class cache..."); //$NON-NLS-1$
					cache = (Vector<CommandReplyPair>)oStream.readObject();
					oStream.close();
					bStream.close();
				}
				catch(FileNotFoundException exxy){
				}
				catch(Exception exxy){
					logr.logError(JDILogger.LEVEL_VERBOSE, "Error reading class cache.  See stack trace."); //$NON-NLS-1$
					exxy.printStackTrace(new PrintStream(logr.getErrorStream()));
				}
			}
			if (cache == null){
				cache = new Vector<CommandReplyPair>();
			}
			logr.log(JDILogger.LEVEL_VERBOSE, cache.size() + " objects cached"); //$NON-NLS-1$
		}

		//Search cache
		for(int i = 0; i < cache.size(); i++){
			CommandReplyPair crp = cache.elementAt(i);
			CommandPacket cp = crp.getCommandPacket();
			if (cp.getCommandSet() == cmdpckt.getCommandSet()){
				if (cp.getCommand() == cmdpckt.getCommand()){
					if (Arrays.equals(cp.getByteData(), cmdpckt.getByteData())){
						ReplyPacket rp = crp.getReplyPacket();
						rp.setSequence(cmdpckt.getSequence());
						//Make sure this packet gets to the destination
						rp.setVSock(cmdpckt.getVSock());
						return rp;
					}
				}
			}
		}
		return null;
	}


	public synchronized ReplyPacket query(CommandPacket cmdpckt){

		if (this.cacheSettings != CACHE_NOCACHE){
			ReplyPacket rply = readCache(cmdpckt);
			if (rply != null){
				return rply;
			}


		}

		long ctime = System.currentTimeMillis();

		ReplyPacket rpckt = null;
		try{
			rpckt = (ReplyPacket)provider.query(cmdpckt);
		}catch(Exception exxy){
			logr.logError(JDILogger.LEVEL_NORMAL, "Uncaught exception"); //$NON-NLS-1$
			exxy.printStackTrace(new PrintStream(logr.getErrorStream()));
			rpckt = new ReplyPacket(cmdpckt.getSequence(), KatoReader.FLAG_REPLY_PACKET, KatoReader.ERROR_INTERNAL);
			return rpckt;
		}
		ctime = System.currentTimeMillis() - ctime;
		//This skips cache stuff if we turned off caching here
		//And if the packet is null
		if (this.cacheSettings == CACHE_NOCACHE || rpckt == null){
			return rpckt;
		}

		//If the reply packet didn't specify a cache setting
		if (rpckt.getCacheSetting() == Packet.CACHE_UNSET){
			//Use the command packet's
			rpckt.setCacheSetting(cmdpckt.getCacheSetting());
		}

		//If we're forcing a cache, or was asked to cache, or the packet level cache request was unset and the cache is on
		if (this.cacheSettings == CACHE_FORCECACHE || rpckt.getCacheSetting() == Packet.CACHE_CACHE || (this.cacheSettings == CACHE_CACHE && rpckt.getCacheSetting() == Packet.CACHE_UNSET)){
			//If the time was too great and explicit cache is off (requires packets to ask for cache),
			//We're forcing cache, or the packet requested to be cached.
			if ((ctime >= cacheTime && this.cacheSettings != CACHE_EXPLICIT) || this.cacheSettings == CACHE_FORCECACHE || rpckt.getCacheSetting() == Packet.CACHE_CACHE){
				cachePacket(cmdpckt, rpckt);
				


			}

		}

		return rpckt;
	}
	
	private void cachePacket(CommandPacket cmdpckt, ReplyPacket rpckt){
		logr.log(JDILogger.LEVEL_VERYVERBOSE, "Cacheing packet"); //$NON-NLS-1$
		for(int i = 0; i < cache.size(); i++){
			CommandReplyPair crp = cache.elementAt(i);
			CommandPacket cp = crp.getCommandPacket();
			if (cp.getCommandSet() == cmdpckt.getCommandSet()){
				if (cp.getCommand() == cmdpckt.getCommand()){
					if (Arrays.equals(cp.getByteData(), cmdpckt.getByteData())){
						//This means we already have it, don't add it again
						return;
					}
				}
			}

		}
		CommandReplyPair ncrp = new CommandReplyPair(cmdpckt, rpckt);
		cache.add(ncrp);
		if (!doDelay){
			resaveCache();
		}
	}

	public void tweakCacheTime(int inTime){
		cacheTime = inTime;
	}

	private void resaveCache(){
		if (fileSettings == FILE_USEFILE && cache != null){
			try{
				logr.log(JDILogger.LEVEL_VERBOSE, "Writing " + cache.size() + " packets to file cache"); //$NON-NLS-1$ //$NON-NLS-2$
				FileOutputStream bStream = new FileOutputStream(hashCode + ".dat"); //$NON-NLS-1$
				ObjectOutputStream oStream = new ObjectOutputStream(bStream);
				oStream.writeObject(cache);
				oStream.close();
				bStream.close();
			}catch(Exception exxy){
				exxy.printStackTrace();
			}
		}
	}
	public String getHash(){
		if (this.hashCode != null){
			return this.hashCode;
		}

		String fullHash;
		if (disableHash || fileSettings != FILE_USEFILE || cacheSettings == CACHE_NOCACHE){
			return "JDICache"; //$NON-NLS-1$
		}
		try{
			MessageDigest md5er = MessageDigest.getInstance("md5"); //$NON-NLS-1$
			try{
				FileInputStream fi = new FileInputStream(coreFile);
				while (fi.available() > 0){
					byte []buff = new byte[256];
					fi.read(buff);
					md5er.update(buff);
				}
				fi.close();
				byte []digest = md5er.digest();

				StringBuffer strbld = new StringBuffer();
				for(int i = 0; i < digest.length; i++){
					String temp = Integer.toHexString(digest[i] & 0xFF);
					if (temp.length() < 2){
						temp = "0" + temp; //$NON-NLS-1$
					}
					strbld.append(temp);
				}
				fullHash = strbld.toString();
			}
			catch(IOException exxy){
				JDILogger.logger.logError(JDILogger.LEVEL_NORMAL, "Problem processing hash for core file"); //$NON-NLS-1$
				fullHash = "JDICache"; //$NON-NLS-1$
			}

		}catch(NoSuchAlgorithmException exxy){
			JDILogger.logger.logError(JDILogger.LEVEL_NORMAL, "No MD5 Algorithm"); //$NON-NLS-1$
			fullHash = "JDICache"; //$NON-NLS-1$
		}
		logr.log(JDILogger.LEVEL_VERYVERBOSE, "Dump hash: " + fullHash); //$NON-NLS-1$
		return fullHash;

	}

	public CacherProvider getProvider(){
		return provider;
	}



}
