/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.apache.kato.tools.jdi;
import java.util.concurrent.CountDownLatch;

public class JDIController {
	
	private static CountDownLatch readyToGo;
	
	public static void jdbLaunch(String args[], CountDownLatch readyToGo){
		JDIController.readyToGo = readyToGo;
		JDIController.main(args);
	}

	public static void main(String args[]){
		long baselineMemory = Runtime.getRuntime().totalMemory() - Runtime.getRuntime().freeMemory();

		int cacheTime = -1;
		boolean useFiles = true;
		boolean delayWrite = true;
		boolean disableHash = false;
		boolean autoTerm = false;
		int port = 8000;


		JDILogger.logger = new JDILogger();
		JDILogger logr = JDILogger.logger;

		String coreFile = ""; //$NON-NLS-1$
		if (args.length == 1){
			coreFile = args[0];
		}else if (args.length > 0){
			try{
				for(int i = 0; i < args.length; i++){
					if (args[i].equalsIgnoreCase("-v")){ //$NON-NLS-1$
						logr.setVerbosity(logr.getVerbosity() + 1);
					}else if (args[i].equalsIgnoreCase("-q")){ //$NON-NLS-1$
						logr.setVerbosity(logr.getVerbosity() - 1);
					}else if (args[i].equalsIgnoreCase("--nofile")){ //$NON-NLS-1$
						useFiles = false;
					}else if (args[i].equalsIgnoreCase("-a")){ //$NON-NLS-1$
						autoTerm = true;
					}else if (args[i].equalsIgnoreCase("-p")){ //$NON-NLS-1$
						try{
							port = Integer.parseInt(args[i + 1]);
							i++;
						}catch(NumberFormatException exxy){
							System.err.println("Port number (-p) is not a number, or malformed"); //$NON-NLS-1$
							printUsage();
							System.exit(1);
						}
					}else if (args[i].equalsIgnoreCase("--nohash")){ //$NON-NLS-1$
						disableHash = true;
					}else if (args[i].equalsIgnoreCase("--delaywrite")){ //$NON-NLS-1$
						delayWrite = true;
					}else if (args[i].equalsIgnoreCase("-t")){ //$NON-NLS-1$
						try{
							cacheTime = Integer.parseInt(args[i + 1]);
							i++;
						}catch(NumberFormatException exxy){
							System.err.println("Cache time (-t) is not a number, or malformed"); //$NON-NLS-1$
							printUsage();
							System.exit(1);
						}
					}else if (args[i].equalsIgnoreCase("--nocache")){ //$NON-NLS-1$
						cacheTime = -2;
					}
				}
				coreFile = args[args.length - 1];
			}catch(Exception exxy){
				System.err.println("Malformed command line"); //$NON-NLS-1$
				printUsage();
				System.exit(1);
			}

		}else{
			printUsage();
			System.exit(1);
		}

		//Used for loading times
		long time = System.currentTimeMillis();

		//Useful information
		logr.log(JDILogger.LEVEL_VERBOSE, "# Procs:" + Runtime.getRuntime().availableProcessors()); //$NON-NLS-1$
		logr.log(JDILogger.LEVEL_VERBOSE, "Bytes free: " + Runtime.getRuntime().freeMemory()); //$NON-NLS-1$
		logr.log(JDILogger.LEVEL_VERBOSE, "Bytes total: " + Runtime.getRuntime().totalMemory()); //$NON-NLS-1$

		//The JDWP server converts:
		//bytes <---> Packets
		JDWPServer svr = null;
		try{
			svr = new JDWPServer(JDILogger.logger, port);
			if (readyToGo!=null)svr.setTerminateLastClient(true);
		}catch(Exception exxy){
			System.err.println("Unrecoverable error:"); //$NON-NLS-1$
			exxy.printStackTrace();
			System.exit(1);
		}
		

		//Intermediate Cache
		JDICacher chsr = null;
		try{
			chsr = new JDICacher(coreFile, svr, JDILogger.logger);
		}catch(Exception exxy){
			System.err.println("Unrecoverable error:"); //$NON-NLS-1$
			exxy.printStackTrace();
			System.exit(1);
		}
		if (chsr == null || svr == null){
			System.err.println("Unrecoverable error, please check previous errors (and/or turn up verbosity)"); //$NON-NLS-1$
			System.exit(1);
			return;
		}
		
		chsr.disableHash(disableHash);
		chsr.delayedWrites(delayWrite);
		if (!useFiles){
			chsr.setFileAccess(JDICacher.FILE_NOFILE);
		}
		
		if (cacheTime == -2){
			chsr.setCacheSettings(JDICacher.CACHE_NOCACHE);
		}else if (cacheTime >= 0){
			chsr.tweakCacheTime(cacheTime);	
		}else{
			chsr.tweakCacheTime(1000);
		}

		svr.setTerminateLastClient(autoTerm);
		
		//Rather than keep making the same requests to the Kato reader
		//Use an intermediate cache
		svr.setCache(chsr);

		//Save the old settings
		svr.askPacket(new CommandPacket(0, 0, 1, 3, Packet.CACHE_CACHE));


		long memory = Runtime.getRuntime().totalMemory() - Runtime.getRuntime().freeMemory();
		logr.log(JDILogger.LEVEL_VERYVERBOSE, "Loading took " + (System.currentTimeMillis() - time) + "ms (" + (memory - baselineMemory) + "B used)"); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$

		//Start the JDWP server.  This call blocks until the JDISocketServer is done.
		//This also starts the JDISocketServer
		try{
			svr.start();
		}catch(Exception exxy){
			
		}
		
		
		memory = Runtime.getRuntime().totalMemory() - Runtime.getRuntime().freeMemory();
		chsr.commit();
		// Notify jdb connector if it is listening...
		if (readyToGo != null){
			readyToGo.countDown();
		}
		logr.log(JDILogger.LEVEL_VERBOSE, "Goodbye! (Uptime: " + (System.currentTimeMillis() - time) + "ms) (" + (memory - baselineMemory) + "B used)"); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
		try{Thread.sleep(5000);}catch(Exception exxy){}
	}

	public static void printUsage(){
		System.out.println(" JDI Command Line Post-Mortem Debugger"); //$NON-NLS-1$
		System.out.println("------------------------------------------"); //$NON-NLS-1$
		System.out.println("-v           Increase verbosity (for each instance)"); //$NON-NLS-1$
		System.out.println("-q           Decrease verbosity (for each instance)"); //$NON-NLS-1$
		System.out.println("-t           Set cache time"); //$NON-NLS-1$
		System.out.println("-a           Terminate server after last client disconnects"); //$NON-NLS-1$
		System.out.println("-p           TCPIP port number."); //$NON-NLS-1$
		System.out.println("--nohash     Do not MD5 hash core. (!USE WITH CAUTION!)*"); //$NON-NLS-1$
		System.out.println("--nofile     File cache is no loaded or saved, caches still happen*"); //$NON-NLS-1$
		System.out.println("--delaywrite Resave cache file only on shutdown*"); //$NON-NLS-1$
		System.out.println("--nocache    Disable cache"); //$NON-NLS-1$
		System.out.println("Last argument is path to dump file"); //$NON-NLS-1$
		System.out.println("Default TCPIP port is 8000."); //$NON-NLS-1$
		System.out.println("* May improve speed over a network"); //$NON-NLS-1$

	}
}
