/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.apache.kato.tools.jdi;


import org.apache.kato.tools.jdi.JDINotifier;

/** JDIServerContainer
 * 
 * The JDIServerContainer is a thread that launches the JDI stuff.  It provides a status back to
 * the JDIServerController of it's status.  Since JDWPServer.start() waits until the VadderSocketServer
 * has stopped running, we need to hold it inside it's own thread.  We provide a pass through to allow the
 * controller to shut down the thread (kill()), and grab any errors (getLastException()).  JDWPServer.stop()
 * calls through to the JDISocketServer and manually shuts it down.  On shut down, the socket server closes
 * all known connections, and returns.  The join() with the server thread eventually returns into this thread,
 * which quits.
 */


public class JDIServerContainer extends Thread {
	private JDWPServer svr;
	private JDICacher chsr;
	private boolean isRunning = false;
	private Exception exxy;
	private JDINotifier notifier = null;
	private boolean hasStarted = false;

	public boolean isRunning(){
		return isRunning;
	}


	public JDWPServer getJDWPServer(){
		return svr;
	}

	public JDIServerContainer(JDWPServer svrin, JDICacher chsrin, JDINotifier not){
		svr = svrin;
		chsr = chsrin;
		notifier = not;
	}

	@Override
	public void start(){
		//Reset exception
		exxy = null;
		isRunning = true;
		//Start the socket stuff
		try{
			svr.start();

			//Now start the thread
			super.start();

			//Wait for thread to start
			while(!hasStarted()){}
		}catch(Exception exxy){
			isRunning = false;
		}
	}

	public void	kill(){
		if (isRunning){
			svr.stop();
			//If we have a notifier, tell them we've stopped running
			if (notifier != null){
				notifier.update(this);
			}

			//Wait to shut down
			while(isRunning){}
		}
	}

	public Exception getLastException(){
		return exxy;
	}

	public boolean hasStarted(){
		return hasStarted;
	}

	@Override
	public void run(){
		//If we have a notifier, tell them we've stopped running
		if (notifier != null){
			notifier.update(this);
		}
		hasStarted = true;

		try{
			//Start the JDWPServer
			svr.getSocketServer().join();
		}catch(Exception lexxy){
			svr.stop();
		}

		//Make sure the cache gets written
		chsr.commit();
		chsr = null;
		svr = null;
		isRunning = false;
		//If we have a notifier, tell them we've stopped running
		if (notifier != null){
			notifier.update(this);
		}
	}
}
