/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.apache.kato.tools.jdi;

import java.lang.ref.WeakReference;

/* Packet
 * 
 * This is the superclass of both the CommandPacket and ReplyPacket, and implements
 * common data information between them.
 * 
 * We maintain a WeakReference to our parent packet (original JDISocket).  At the
 * JDWPServer level, we set the packet, and then send it down to Kato, then set the
 * convert to bytes and send it to the socket.  Most of the time, nothing below the
 * JDWPServer sets a reply packet socket, so we do it in JDWPServer if it's unset.
 * 
 * JDICacher sometimes sets the socket explicitly.
 * 
 * We also keep a packet level cache setting, which may or may not be set by the
 * KatoReader, however, the JDICacher sometime may override with the ForceCache
 * option.
 */

public abstract class Packet implements java.io.Serializable{

	private static final long serialVersionUID = 2173969485927382589L;

	public abstract byte[] toBytes();

	public static final int CACHE_UNSET = 0;
	public static final int CACHE_CACHE = 1;
	public static final int CACHE_NOCACHE = 2;

	public final int DATA_STRING = 1;
	public final int DATA_BYTE = 2;

	protected transient WeakReference<JDISocket> vsock;
	protected int cacheSettings = CACHE_UNSET;
	protected int dataMode;
	protected transient int nSeq;
	protected int nFlags;
	protected byte bdata[];
	
	public Packet(){
		bdata = new byte[0];
	}
	public int getSequence(){
		return nSeq;
	}

	public int getFlags(){
		return nFlags;
	}

	public int getSize(){
		return 11 + bdata.length;
	}

	public void setData(String szData){
		//Set flag that we were set with string data
		dataMode = DATA_STRING;
		bdata = szData.getBytes();
	}

	public void setData(byte[] szData){
		//Set flag that we were set with byte data
		dataMode = DATA_BYTE;
		bdata = szData;
	}

	public String getStringData(){
		if (dataMode == DATA_STRING){
			return bdata.toString();
		}
		throw new ArrayStoreException("The data was set with a byte string"); //$NON-NLS-1$

	}

	public void setVSock(JDISocket ivsock){
		vsock = new WeakReference<JDISocket>(ivsock);
	}

	public JDISocket getVSock(){
		if (vsock == null){
			vsock = new WeakReference<JDISocket>(null);
		}
		return vsock.get();
	}

	public void setCacheSetting(int inCache){
		cacheSettings = inCache;
	}

	public int getCacheSetting(){
		return cacheSettings;
	}

	public byte[] getByteData(){
		return bdata;
	}


}
