/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.apache.kato.tests.junit;

import java.util.Iterator;

import javax.tools.diagnostics.image.CorruptDataException;
import javax.tools.diagnostics.image.DataUnavailable;
import javax.tools.diagnostics.image.ImageSection;
import javax.tools.diagnostics.runtime.java.JavaField;
import javax.tools.diagnostics.runtime.java.JavaMethod;


public class JavaMethodTest extends AbstractImageTestcase
{
	private JavaMethod _method = null;
	
	protected void setUp() throws Exception
	{
		_method = defaultJavaMethod();
		super.setUp();
	}

	/**
	 * Test method for 'javax.tools.diagnostics.runtime.java.JavaMethod.getBytecodeSections()'
	 * 
	 * Ensures that a non-null iterator is returned and every element of the iterator is ImageSection.
	 * (note that the iterator can be empty in the case of natives)
	 */
	public void testGetBytecodeSections()
	{
		Iterator it = _method.getBytecodeSections().iterator();
		assertNotNull(it);
		while (it.hasNext()) {
			assertTrue(it.next() instanceof ImageSection);
		}
	}

	/**
	 * Test method for 'javax.tools.diagnostics.runtime.java.JavaMethod.getCompiledSections()'
	 * 
	 * Ensures that a non-null, non-empty iterator is returned
	 */
	public void testGetCompiledSections()
	{
		Iterator it = _method.getCompiledSections().iterator();
		assertNotNull(it);
		assertTrue(it.hasNext());
	}

	/**
	 * Test method for 'javax.tools.diagnostics.runtime.java.JavaMember.getModifiers()'
	 * 
	 * Ensures that the call succeeds
	 */
	public void testGetModifiers()
	{
		try {
			_method.getModifiers();
		} catch (CorruptDataException e) {
		}
	}

	/**
	 * Test method for 'javax.tools.diagnostics.runtime.java.JavaMember.getDeclaringClass()'
	 * 
	 * Ensures that the class has a non-null address
	 */
	public void testGetDeclaringClass()
	{
		try {
			assertTrue(0 != _method.getDeclaringClass().getID().getAddress());
		} catch (CorruptDataException e) {
		} catch (DataUnavailable e) {
		}
	}

	/**
	 * Test method for 'javax.tools.diagnostics.runtime.java.JavaMember.getName()'
	 * 
	 * Ensure that the method has a non-empty name
	 */
	public void testGetName()
	{
		try {
			assertTrue(_method.getName().length() > 0);
		} catch (CorruptDataException e) {
		}
	}

	/**
	 * Test method for 'javax.tools.diagnostics.runtime.java.JavaMember.getSignature()'
	 * 
	 * Ensure that the signature is non-empty
	 */
	public void testGetSignature()
	{
		try {
			assertTrue(_method.getSignature().length() > 0);
		} catch (CorruptDataException e) {
		}
	}
	
	/**
	 * Verify that the equals call doesn't throw
	 */
	public void testEquals()
	{
		try{
			assertTrue(_method.equals(_method));
		} catch (Throwable t) {
			assertTrue(false);
		}
	}
	
	/**
	 * Verify that hashCode() doesn't throw and returns non-zero (technically zero is ok but it will be 
	 * flagged here to ensure that we aren't doing anything bad to create the hashcode)
	 */
	public void testHashCode()
	{
		try {
			assertTrue(0 != _method.hashCode());
		} catch (Throwable t) {
			assertTrue(false);
		}
	}

	public void testToString()
	{
		String theString=_method.toString();
		assertNotNull(theString);
		assertTrue(theString.length() >= 0 );
	}

}
