/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.apache.kato.tests.junit;

import java.util.Iterator;

import javax.tools.diagnostics.image.CorruptDataException;
import javax.tools.diagnostics.image.DataUnavailable;
import javax.tools.diagnostics.runtime.java.JavaThread;


public class JavaThreadTest extends AbstractImageTestcase
{
	private JavaThread _thread = null;
	private JavaThread _sameThread = null;
	
	protected void setUp() throws Exception
	{
		_thread = defaultJavaThread();
		_sameThread= defaultJavaThread();
		assertNotNull(_thread);
		super.setUp();
	}

	/**
	 * Test method for 'javax.tools.diagnostics.runtime.java.JavaThread.getJNIEnv()'
	 * 
	 * Ensures that the JNIEnv is non-null
	 */
	public void testGetJNIEnv()
	{
		try {
			assertNotNull(_thread.getJNIEnv());
		} catch (CorruptDataException e) {
			//this is acceptable by the spec
		}
	}

	/*
	 * Test method for 'javax.tools.diagnostics.runtime.java.JavaThread.getPriority()'
	 */
	public void testGetPriority()
	{
		try {
			int priority = _thread.getPriority();
			assertTrue((priority >= 1) && (priority <= 10));
		} catch (CorruptDataException e) {
			//this is acceptable by the spec
		}		
	}

	/*
	 * Test method for 'javax.tools.diagnostics.runtime.java.JavaThread.getObject()'
	 */
	public void testGetObject()
	{
		//anything returned here is legal.  In fact, even null is ok
		try {
			_thread.getObject();
		} catch (CorruptDataException e) {
			//this is acceptable by the spec
		}
	}

	/*
	 * Test method for 'javax.tools.diagnostics.runtime.java.JavaThread.getState()'
	 */
	public void testGetState()
	{
		try {
			assertTrue(_thread.getState() > 0);
		} catch (CorruptDataException e) {
			//this is acceptable by the spec
		}
	}

	/**
	 * Test method for 'javax.tools.diagnostics.runtime.java.JavaThread.getImageThread()'
	 * 
	 * Ensures that a non-null image thread can be pulled from this thread
	 */
	public void testGetImageThread()
	{
		try {
			assertNotNull(_thread.getImageThread());
		} catch (CorruptDataException e) {
			//this is acceptable by the spec
		} catch (DataUnavailable e) {
			//this is acceptable by the spec
		}
	}

	/**
	 * Test method for 'javax.tools.diagnostics.runtime.java.JavaThread.getStackSections()'
	 * 
	 * Ensures that a non-null, non-empty iterator is returned
	 */
	public void testGetStackSections()
	{
		Iterator it = _thread.getStackSections().iterator();
		assertNotNull(it);
		assertTrue(it.hasNext());
	}

	/**
	 * Test method for 'javax.tools.diagnostics.runtime.java.JavaThread.getStackFrames()'
	 * 
	 * Ensures that the frame iterator is non-null even though it may be empty
	 */
	public void testGetStackFrames()
	{
		Iterator frames = _thread.getStackFrames().iterator();
		assertNotNull(frames);
		//it is legal for there to be no frames
	}

	/*
	 * Test method for 'javax.tools.diagnostics.runtime.java.JavaThread.getName()'
	 */
	public void testGetName()
	{
		try {
			assertNotNull(_thread.getName());
		} catch (CorruptDataException e) {
			//this is acceptable by the spec
		}
	}

	/**
	 * Verify that the equals call doesn't throw
	 */
	public void testEquals()
	{
		try{
			assertTrue(_thread.equals(_thread));
			assertTrue(_thread.equals(_sameThread));
		} catch (Throwable t) {
			assertTrue(false);
		}
	}
	
	/**
	 * Verify that hashCode() doesn't throw and returns non-zero (technically zero is ok but it will be 
	 * flagged here to ensure that we aren't doing anything bad to create the hashcode)
	 */
	public void testHashCode()
	{
		try {
			assertTrue(0 != _thread.hashCode());
			assertTrue(_thread.hashCode() == _sameThread.hashCode());
		} catch (Throwable t) {
			assertTrue(false);
		}
	}
	
	/**
	 * A test to make sure that all the threads in the runtime have a name and priority.
	 * 
	 * This exists to prove that we can get the name and priority even in cases where the threads are subclasses of java/lang/Thread
	 */
	public void testAllThreadsNamed()
	{
		Iterator threads = defaultJavaRuntime().getThreads().iterator();
		
		while (threads.hasNext()) {
			JavaThread thread = (JavaThread) threads.next();
			
			try {
				assertNotNull(thread.getName());
				thread.getPriority();
			} catch (CorruptDataException e) {
				assertTrue(false);
			}
		}
	}
}
