/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.apache.kato.katoview.commands;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Stack;

import javax.tools.diagnostics.image.CorruptDataException;
import javax.tools.diagnostics.image.DataUnavailable;
import javax.tools.diagnostics.image.Image;
import javax.tools.diagnostics.image.ImageSection;
import javax.tools.diagnostics.runtime.ManagedRuntime;
import javax.tools.diagnostics.runtime.java.JavaClass;
import javax.tools.diagnostics.runtime.java.JavaHeap;
import javax.tools.diagnostics.runtime.java.JavaMethod;
import javax.tools.diagnostics.runtime.java.JavaObject;
import javax.tools.diagnostics.runtime.java.JavaRuntime;

import org.apache.kato.katoview.Output;
import org.apache.kato.katoview.commands.helpers.Utils;


public class WhatisCommand extends Command {

	public WhatisCommand(Output o){
		super(o, "whatis", "gives information about what is stored at the given memory address",
				"parameters: <hex_address>\n\n" +
				"the whatis command examines the memory location at <hex_address> and tries to find out more information about this address " +
				"- for example whether it is within an object in a heap or within the byte codes associated with a class method."
				);
		child_commands = null;
	}
	
	public void doCommand(Stack args, Image loadedImage, HashMap systemProperties){
		if(0 == args.size()){
			out.error("missing address argument");
			return;
		}
		String address = (String)args.pop();
		Long lAddressInDecimal = Utils.longFromString(address);
		if (null == lAddressInDecimal){
			out.error("cannot convert \"" + address + "\" to numeric value"); 
			return;
		}
		long addressInDecimal = lAddressInDecimal.longValue();
		
		findInImage(loadedImage, addressInDecimal);
		
		out.print("\n");
	}
	
	private void findInImage(Image loadedImage, long address)
	{
		ManagedRuntime mr;
		Iterator itRuntime = Utils.getRuntimes(loadedImage);
		int count = 1;

		out.print("\n");
		while (itRuntime.hasNext()) {
			mr = (ManagedRuntime)itRuntime.next();
			if (mr instanceof JavaRuntime)
			{
				out.print("\truntime #" + count + "\n");
				
				findInRuntime((JavaRuntime)mr, address);
			}
			count++;
		}
	}
	
	private void findInRuntime(JavaRuntime jr, long address)
	{
		JavaHeap jh;
		Iterator itHeap = jr.getHeaps().iterator();
		int count = 1;
		
		while (itHeap.hasNext()) {
			jh = (JavaHeap)itHeap.next();
			
			out.print("\t  heap #" + count + " - name: ");
			out.print(jh.getName() + "\n");
			
			findInHeap(jh, address);
			count++;
		}
	}
	
	private void findInHeap(JavaHeap jh, long address)
	{
		//TODO: checkWithinValidMemRange(sb, ...);
		
		if (isWithinImageSections(jh.getSections().iterator(), null, false, address)){
			//if it's start or within the range of an object
			if (!isStartOfObj(jh.getObjects().iterator(), address)){
				if (!isWithinObjectRange(jh.getObjects().iterator(), address)){
					out.print("\t\t0x" + Long.toHexString(address) + " is orphaned on the heap.\n");
				}
			}
		}
		else{
			out.print("\t\t0x" + Long.toHexString(address) + " is not within this heap.\n");
			//TODO : function to indicate 16 or 32 bit 
			long bound = 12; //bounds default to 16 for 32 bit system.
			checkClassInRange(jh.getObjects().iterator(), bound, address);
		}
		
		checkMethodInRange(jh.getObjects().iterator(), address);
	}
	
	
	
	private void checkMethodInRange(Iterator objects, long address){
		while(objects.hasNext()){
			JavaObject jObject = (JavaObject)objects.next();
			JavaClass jClass;
			try{
				jClass = jObject.getJavaClass();
			}catch(CorruptDataException cde){
				//go to the next iteration
				continue;
			}
			Iterator methods = jClass.getDeclaredMethods().iterator();
			while(methods.hasNext()){
				JavaMethod jMethod = (JavaMethod)methods.next();
				Iterator bytecodeSections = jMethod.getBytecodeSections().iterator();
				Iterator compiledSections = jMethod.getCompiledSections().iterator();
				isWithinImageSections(bytecodeSections, jMethod, false, address);
				isWithinImageSections(compiledSections, jMethod, true, address);
			}
		}
	}

/*	TODO: need to implement some output for this method and use it
	private void checkMonitorInRange(Iterator monitors){
		while(monitors.hasNext()){
			JavaMonitor jMonitor = (JavaMonitor)monitors.next();
			//TODO: need API for retrieving monitor ID (address)
			//sb.append("\n" + "monitor object id: " + jMonitor.getObject().getID().getAddress());
		}
	}
*/
	
	private void checkClassInRange(Iterator objects, long bound, long address){
		long startAddress, endAddress ;
		while(objects.hasNext()){
			JavaObject jObject = (JavaObject)objects.next();
			JavaClass jClass;
			String className;
			try{
				jClass = jObject.getJavaClass();
				className = jClass.getName();
			}catch(CorruptDataException cde){
				//TODO exception handling
				continue;
			}
			startAddress = jClass.getID().getAddress();
			endAddress = startAddress + bound;
			if(address == startAddress){
				out.print("0x" + Long.toHexString(address) +
						" is the address of the java/lang/Class object for " + className);
				return;
			}
			if (isWithinRange(startAddress, endAddress, address)){
				out.print("0x" + Long.toHexString(address) 
						+ " is within the java/lang/Class object for " + className);
				return;
			}
		}
	}
	
	private boolean isWithinObjectRange(Iterator objects, long address){
		long startAddress, endAddress ;
		String className;
		while(objects.hasNext()){
			JavaObject jObject = (JavaObject)objects.next();
			try{
				className = jObject.getJavaClass().getName();
				startAddress = jObject.getID().getAddress();
				endAddress = startAddress + jObject.getSize();
			}catch(CorruptDataException cde){
				//TODO exception handling
				continue;
			}
			if (isWithinRange(startAddress, endAddress, address)){
				out.print("\t\t0x" + Long.toHexString(address) + " is within an object on the heap:\n" + 
						"\t\t\toffset " + (address - startAddress) + " within "+ className +
						" instance @ 0x" + Long.toHexString(startAddress) + "\n");
				return true;
			}
		}
		return false;
	}
	
	private boolean isWithinRange(long startAddress, long endAddress, long address){
		return (address <= endAddress && address > startAddress);
	}
	
	private boolean isStartOfObj(Iterator objects, long address){
		String className;
		while(objects.hasNext()){
			JavaObject jObject = (JavaObject)objects.next();
			if (address == jObject.getID().getAddress()){
				try{
					className = jObject.getJavaClass().getName();
				}catch(CorruptDataException cde){
					className = "N/A (CorruptDataException occurred)";
				}
				out.print("0x" + Long.toHexString(address) + " is start of an object of type " + className);
				return true;
			}
		}
		return false;
	}
	
	private boolean isWithinImageSections(Iterator heapImageSections, Object memType,
	  boolean isMethodCompiled, long address)
	{
		while (heapImageSections.hasNext()){
			ImageSection imageSection = (ImageSection)heapImageSections.next();
			long baseAddress = imageSection.getBaseAddress().getAddress();
			long size = imageSection.getSize();
			long endAddress = baseAddress + size;
			
			if (address <= endAddress  && address >= baseAddress) {
				if (null == memType) {
					out.print("\t\t0x" + Long.toHexString(address) + " is within heap segment: " +  
							Long.toHexString(baseAddress) + " -- " + Long.toHexString(endAddress) + "\n");
					return true;
				}
				if (memType instanceof JavaMethod) {
					String methodName = "N/A", methodSig = "N/A", className = "N/A";
					try{
						methodName = ((JavaMethod)memType).getName();
						methodSig = ((JavaMethod)memType).getSignature();
						className = ((JavaMethod)memType).getDeclaringClass().getName();
					}catch(CorruptDataException cde){				
					}catch(DataUnavailable du){
					}
					String codeType = isMethodCompiled ? "compiled code" : "byte code";
					out.print("0x" + Long.toHexString(address) + " is within the " + codeType + " range: " +  
							Long.toHexString(baseAddress) + " -- " + Long.toHexString(endAddress) + "\n\t" +  "...of method " +
							methodName + " with signature " + methodSig + "\n\t" + 
							"...in class " + className + "\n");
					return true;
				}
			}
		}
		return false;
	}
}
