/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.apache.kato.katoview.heapdump;

import java.io.IOException;

/**
 * Abstract class to decouple the formatting of heapdumps from the walking of
 * the internal data structures that hold the source data.
 * 
 */
public abstract class HeapDumpFormatter
{
	protected final String _version;   
	protected final boolean _is64Bit;

	/**
	 * Constructor
	 * 
	 * @param version JVM version information to be included in heapdump header
	 * @param is64Bit True if we are dumping a 64 bit image, false otherwise
	 */
	protected HeapDumpFormatter(String version,boolean is64Bit)
	{
		this._version = version;
		this._is64Bit = is64Bit;
	}

	/**
	 * Adds a class to the heapdump.
	 * 
	 * @param address
	 *            Address of class
	 * @param name
	 *            Name of class
	 * @param superClassAddress
	 *            Address of superclass
	 * @param size
	 *            Size of class, bytes
	 * @param instanceSize
	 *            Size of objects of this class
	 * @param hashCode
	 *            Hashcode
	 * @param references
	 *            Iterator of Long addresses (some of which may be null)
	 * @throws IOException
	 */
	public abstract void addClass(long address, String name,
			long superClassAddress, int size, int instanceSize, int hashCode, ReferenceIterator references)
			throws IOException;

	/**
	 * Adds an object to the heapdump.
	 * 
	 * @param address
	 *            Address of object
	 * @param classAddress
	 *            Address of class of object
	 * @param className
	 *            Name of class of object
	 * @param size
	 *            Size of object, bytes
	 * @param hashCode
	 *            Hashcode of object
	 * @param references
	 *            Iterator of Long addresses (some of which may be null)
	 * @throws IOException
	 */
	public abstract void addObject(long address, long classAddress,
			String className, int size, int hashCode, ReferenceIterator references)
			throws IOException;

	/**
	 * Adds a primitive array object to the heapdump
	 * 
	 * Type code is: 0 - bool, 1 - char, 2 - float, 3 - double, 4 - byte, 5 - short, 6 -
	 * int, 7 - long
	 * 
	 * @param address
	 *            Address of object
	 * @param arrayClassAddress
	 *           Address of primitive array class
	 * @param type
	 *            Type code - see above
	 * @param size
	 *            Size of object, bytes.
	 * @param hashCode
	 *            Hashcode of object
	 * @param numberOfElements
	 *            Number of elements in the array
	 * @throws IOException
	 * @throws IllegalArgumentException
	 *             if invalid type is supplied
	 */
	public abstract void addPrimitiveArray(long address, long arrayClassAddress, int type, int size,
			int hashCode, int numberOfElements) throws IOException,
			IllegalArgumentException;

	/**
	 * Adds an object array to the heapdump.
	 * 
	 * @param address
	 *            Address of object
	 * @param arrayClassAddress
	 *            Address of class of array object
	 * @param arrayClassName
	 *            Name of class of array object
	 * @param elementClassAddress
	 *            Address of class of elements
	 * @param elementClassName
	 *            Name of class of elements
	 * @param size
	 *            Size of object, bytes
	 * @param numberOfElements
	 *            number of elements in the array
	 * @param hashCode
	 *            Hashcode of object
	 * @param references
	 *            Iterator of Long addresses (some of which may be null)
	 * @throws IOException
	 */
	public abstract void addObjectArray(long address, long arrayClassAddress,
			String arrayClassName, long elementClassAddress,
			String elementClassName, int size, int numberOfElements,
			int hashCode, ReferenceIterator references) throws IOException;
	
	/**
	 * Closes the heapdump
	 * 
	 * @throws IOException
	 */
	public abstract void close() throws IOException;
}
